(* Read and write document summary information

    Dr. J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Jul. 2009
         2 - Jan. 2016 : text moved to "resourcestring" in unit "FilePropConsts"
         2.1 - July 2021 : LoadLibrary replaced by SafeLoadLibrary to preserve
                the current FPU control word

   last updated: July 2021
   *)

unit FilePropUtils;

interface

uses
  Winapi.Windows, System.Win.ComObj, Winapi.ActiveX;

const
  FMTID_SummaryInformation          : TGUID =
    '{F29F85E0-4FF9-1068-AB91-08002B27B3D9}';
  FMTID_DocSummaryInformation       : TGUID =
    '{D5CDD502-2E9C-101B-9397-08002B2CF9AE}';
  FMTID_UserDefinedProperties       : TGUID =
    '{D5CDD505-2E9C-101B-9397-08002B2CF9AE}';
  FMTID_AudioSummaryInformation     : TGuid =
    '{64440490-4C8B-11D1-8B70-080036B11A03}';
  FMTID_VideoSummaryInformation     : TGUID =
    '{64440491-4C8B-11D1-8B70-080036B11A03}';
  FMTID_ImageSummaryInformation     : TGUID =
    '{6444048f-4c8b-11d1-8b70-080036b11a03}';
  FMTID_MediaFileSummaryInformation : TGUID =
    '{64440492-4c8b-11d1-8b70-080036b11a03}';


  STGFMT_STORAGE     = 0;
  STGFMT_FILE        = 3;
  STGFMT_ANY         = 4;
  STGFMT_DOCFILE     = 5;

  // Summary Information
  PIDSI_TITLE        = $00000002;     // VT_LPSTR
  PIDSI_SUBJECT      = $00000003;
  PIDSI_AUTHOR       = $00000004;
  PIDSI_KEYWORDS     = $00000005;
  PIDSI_COMMENTS     = $00000006;
  PIDSI_TEMPLATE     = $00000007;
  PIDSI_LASTAUTHOR   = $00000008;
  PIDSI_REVNUMBER    = $00000009;
  PIDSI_EDITTIME     = $0000000A;     // VT_FILETIME (UTC)
  PIDSI_LASTPRINTED  = $0000000B;
  PIDSI_CREATE_DTM   = $0000000C;
  PIDSI_LASTSAVE_DTM = $0000000D;
  PIDSI_PAGECOUNT    = $0000000E;     // VT_I4
  PIDSI_WORDCOUNT    = $0000000F;
  PIDSI_CHARCOUNT    = $00000010;
  PIDSI_THUMBNAIL    = $00000011;     // VT_CF
  PIDSI_APPNAME      = $00000012;     // VT_LPSTR
  PIDSI_DOC_SECURITY = $00000013;     // VT_I4

  // Document Summary Information
  PIDDSI_CATEGORY    = $00000002;     // VT_LPSTR
  PIDDSI_PRESFORMAT  = $00000003;
  PIDDSI_BYTECOUNT   = $00000004;     // VT_I4
  PIDDSI_LINECOUNT   = $00000005;
  PIDDSI_PARCOUNT    = $00000006;
  PIDDSI_SLIDECOUNT  = $00000007;
  PIDDSI_NOTECOUNT   = $00000008;
  PIDDSI_HIDDENCOUNT = $00000009;
  PIDDSI_MMCLIPCOUNT = $0000000A;
  PIDDSI_SCALE       = $0000000B;     // VT_BOOL
  PIDDSI_HEADINGPAIR = $0000000C;     // VT_VARIANT | VT_VECTOR
  PIDDSI_DOCPARTS    = $0000000D;     // VT_LPSTR | VT_VECTOR
  PIDDSI_MANAGER     = $0000000E;     // VT_LPSTR
  PIDDSI_COMPANY     = $0000000F;
  PIDDSI_LINKSDIRTY  = $00000010;     // VT_BOOL

  // FMTID_MediaFileSummaryInfo
  PIDMSI_EDITOR      = $00000002;     // VT_LPWSTR
  PIDMSI_SUPPLIER    = $00000003;
  PIDMSI_SOURCE      = $00000004;
  PIDMSI_SEQUENCE_NO = $00000005;
  PIDMSI_PROJECT     = $00000006;
  PIDMSI_STATUS      = $00000007;     // VT_UI4
  PIDMSI_OWNER       = $00000008;     // VT_LPWSTR
  PIDMSI_RATING      = $00000009;
  PIDMSI_PRODUCTION  = $0000000A;     // VT_FILETIME (UTC)
  PIDMSI_COPYRIGHT   = $0000000B;     // VT_LPWSTR

  // FMTID_AudioSummaryInformation
  PIDASI_FORMAT        = $00000002;   // VT_BSTR
  PIDASI_TIMELENGTH    = $00000003;   // VT_UI4, milliseconds
  PIDASI_AVG_DATA_RATE = $00000004;   // VT_UI4,  Hz
  PIDASI_SAMPLE_RATE   = $00000005;   // VT_UI4,  bits
  PIDASI_SAMPLE_SIZE   = $00000006;   // VT_UI4,  bits
  PIDASI_CHANNEL_COUNT = $00000007;   // VT_UI4
  PIDASI_STREAM_NUMBER = $00000008;   // VT_UI2
  PIDASI_STREAM_NAME   = $00000009;   // VT_LPWSTR
  PIDASI_COMPRESSION   = $0000000A;   // VT_LPWSTR

  // FMTID_VideoSummaryInformation
  PIDVSI_STREAM_NAME   = $00000002;   // "StreamName", VT_LPWSTR
  PIDVSI_FRAME_WIDTH   = $00000003;   // "FrameWidth", VT_UI4
  PIDVSI_FRAME_HEIGHT  = $00000004;   // "FrameHeight", VT_UI4
  PIDVSI_TIMELENGTH    = $00000007;   // "TimeLength", VT_UI4, milliseconds
  PIDVSI_FRAME_COUNT   = $00000005;   // "FrameCount". VT_UI4
  PIDVSI_FRAME_RATE    = $00000006;   // "FrameRate", VT_UI4, frames/millisecond
  PIDVSI_DATA_RATE     = $00000008;   // "DataRate", VT_UI4, bytes/second
  PIDVSI_SAMPLE_SIZE   = $00000009;   // "SampleSize", VT_UI4
  PIDVSI_COMPRESSION   = $0000000A;   // "Compression", VT_LPWSTR
  PIDVSI_STREAM_NUMBER = $0000000B;   // "StreamNumber", VT_UI2

type
  TFileProperty     = record
    PropId  : cardinal;
    PropVariant : TPropVariant;
    end;
  TFilePropertyArray = array of TFileProperty;

  TFilePropertySet = record
    FormatId  : TGUID;
    PropArray : TFilePropertyArray;
    end;
  TFileSummary = array of TFilePropertySet;

function SummaryGuidToString (FormatId : TGUID) : string;
function FilePropertyToString(FormatId : TGUID; PropId : Cardinal): string;
function GetPropertyValueAsString(PropVariant : TPropVariant) : string;

function GetFileSummary(const FileName: WideString; var Summary : TFileSummary) : HResult;
function SetFileSummary(const FileName: WideString; const Summary : TFileSummary) : HResult;

function GetFilePropertySet(const FileName: WideString; FormatId : TGUID) : string;
function GetFileProperty(const FileName: WideString; FormatId : TGUID; PropId : cardinal): string;

{ ------------------------------------------------------------------- }
// This routine only supports pathnames < MAX_PATH (=260)
// Reason: the called function "StgOpenStorageEx" does not support UNC paths
function CopyFileStorage (const SourceName,DestName : string) : integer;


implementation

uses System.SysUtils, WinApiUtils, FilePropConsts;

const
  szOleDll                = 'ole32.dll';
  IID_IPropertySetStorage : TGUID = '{0000013A-0000-0000-C000-000000000046}';
type
  TStgOpenStorageEx       = function(const pwcsName: PWideChar;
    grfMode: DWORD; stgfmt: DWORD; grfAttrs: DWORD; pStgOptions: pointer;
    reserved2: pointer; riid: PGUID; out ppObjectOpen: IStorage): HRESULT;
    stdcall;
var
  StgOpenStorageEx        : TStgOpenStorageEx = nil;
  ole32dll                : dword = 0;
  Win2k                   : boolean = false;

{ ------------------------------------------------------------------- }
// convert Filetime to Delphi time (TDateTime)
function FileTimeToDateTime (ft : TFileTime) : TDateTime;
var
  st : TSystemTime;
  dt : TDateTime;
begin
  FileTimeToSystemTime(ft,st);
  with st do begin
    if TryEncodeDate(wYear,wMonth,wDay,Result) then begin
      if TryEncodeTime(wHour,wMinute,wSecond,wMilliseconds,dt) then Result:=Result+dt;
      end
    else Result:=EncodeDate(1980,1,1);
    end;
  end;


function SummaryGuidToString(FormatId : TGUID) : string;
begin
  if IsEqualGUID(FormatId,FMTID_SummaryInformation) then Result:=rsGlobalInfo
  else if IsEqualGUID(FormatId,FMTID_DocSummaryInformation) then Result:=rsDocInfo
  else if IsEqualGUID(FormatId,FMTID_UserDefinedProperties) then Result:=rsUserInfo
  else if IsEqualGUID(FormatId,FMTID_AudioSummaryInformation)then Result:=rsAudioInfo
  else if IsEqualGUID(FormatId,FMTID_VideoSummaryInformation) then Result:=rsVideoInfo
  else if IsEqualGUID(FormatId,FMTID_ImageSummaryInformation) then Result:=rsImageInfo
  else if IsEqualGUID(FormatId,FMTID_MediaFileSummaryInformation) then Result:=rsmediaInfo
  else Result:=GuidToString(FormatId);
  end;

// FormatIDs supported: FMTID_SummaryInformation and FMTID_DocSummaryInformation
function FilePropertyToString(FormatId : TGUID; PropId : Cardinal): string;
begin
  if IsEqualGUID(FormatId,FMTID_SummaryInformation) then begin
    case PropId of
    PIDSI_TITLE:       Result:=rsTitel;
    PIDSI_SUBJECT:     Result:=rsSubject;
    PIDSI_AUTHOR:      Result:=rsAuthor;
    PIDSI_KEYWORDS:    Result:=rsKeywords;
    PIDSI_COMMENTS:    Result:=rsComments;
    PIDSI_TEMPLATE:    Result:=rsTemplate;
    PIDSI_LASTAUTHOR:  Result:=rsLastSaved;
    PIDSI_REVNUMBER:   Result:=rsRevNumber;
    PIDSI_EDITTIME:    Result:=rsTotalTime;
    PIDSI_LASTPRINTED: Result:=rsLastPrinted;
    PIDSI_CREATE_DTM:  Result:=rsContentCreate;
    PIDSI_LASTSAVE_DTM:Result:=rsDateLastSaved;
    PIDSI_PAGECOUNT:   Result:=rsPages;
    PIDSI_WORDCOUNT:   Result:=rsWordCount;
    PIDSI_CHARCOUNT:   Result:=rsCharCount;
    PIDSI_THUMBNAIL:   Result:=rsThumbnail;
    PIDSI_APPNAME:     Result:=rsProgName;
    PIDSI_DOC_SECURITY:    Result:=rsSecurity;
    else Result:='$'+IntToHex(PropId,8);
      end;
    end
  else if IsEqualGUID(FormatId,FMTID_DocSummaryInformation) then begin
    case PropId of
    PIDDSI_CATEGORY    : Result:=rsCategory;
    PIDDSI_PRESFORMAT  : Result:=rsPresentation;
    PIDDSI_BYTECOUNT   : Result:=rsByteCount;
    PIDDSI_LINECOUNT   : Result:=rsLineCount;
    PIDDSI_PARCOUNT    : Result:=rsParCount;
    PIDDSI_SLIDECOUNT  : Result:=rsSlides;
    PIDDSI_NOTECOUNT   : Result:=rsNotes;
    PIDDSI_HIDDENCOUNT : Result:=rsHidden;
    PIDDSI_MMCLIPCOUNT : Result:=rsClips;
    PIDDSI_SCALE       : Result:=esScale;
    PIDDSI_HEADINGPAIR : Result:=rsHeadPairs;
    PIDDSI_DOCPARTS    : Result:=rsDocParts;
    PIDDSI_MANAGER     : Result:=rsManager;
    PIDDSI_COMPANY     : Result:=rsCompany;
    PIDDSI_LINKSDIRTY  : Result:=rsLinkUpdate;
    else Result:='$'+IntToHex(PropId,8);
      end;
    end
  else Result:='$'+IntToHex(PropId,8);
  end;

function GetPropertyValueAsString(PropVariant : TPropVariant) : string;
begin
  with PropVariant do case vt of
  VT_EMPTY, VT_NULL: Result:='';
  VT_LPSTR:          Result:=pszVal;
  VT_LPWSTR:         Result:=pwszVal;
  VT_BSTR:           Result:=bstrVal;
  VT_ERROR:          Result:=IntToStr(scode);
  VT_I1:             Result:=IntToStr(bVal);  // no cVal
  VT_I2:             Result:=IntToStr(iVal);
  VT_I4, VT_INT:     Result:=IntToStr(lVal);
  VT_I8:             Result:=IntToStr(hVal.QuadPart);
  VT_UI1:            Result:=IntToStr(bVal);
  VT_UI2:            Result:=IntToStr(uiVal);
  VT_UI4, VT_UINT:   Result:=IntToStr(ulVal);
  VT_UI8:            Result:=IntToStr(uhVal.QuadPart);
  VT_R4:             Result:=FloatToStr(fltVal);
  VT_R8:             Result:=FloatToStr(dblVal);
//  VT_CF:             Result:=pclipdata
  VT_DATE:           Result:=DateTimeToStr(date);
  VT_FILETIME:       Result:=DateTimeToStr(FileTimeToDateTime(filetime));
  VT_BOOL:           if bool then Result:=rsYes else Result:=rsNo;
  VT_CLSID:          Result:=GuidToString(puuid^);
  else Result:='?';
    end;
  end;

function GetFileSummary(const FileName: WideString; var Summary : TFileSummary) : HResult;
var
  hr          : HRESULT;
  Stg         : IStorage;
  PropSetStg  : IPropertySetStorage;
  PropSetEnum : IEnumSTATPROPSETSTG;
  PropEnum    : IEnumSTATPROPSTG;
  PropStg     : IPropertyStorage;
  PropSpec    : TPropSpec;
  PropSetStat : TStatPropSetStg;
  PropStat    : TSTATPROPSTG;
  pv          : TPropVariant;
  i,k,n       : integer;
begin
  Result:=S_OK;
  for i:=0 to High(Summary) do Summary[i].PropArray:=nil;
  Summary:=nil;
  // for 9x and NT4 users
  if (not Win2k) or (@StgOpenStorageEx=nil) then
    Result:=StgOpenStorage(pwidechar(FileName),nil,STGM_READ or
      STGM_SHARE_DENY_WRITE,nil,0,Stg)
  // for 2000, XP and higher
  else if(@StgOpenStorageEx <> nil) then
    Result:=StgOpenStorageEx(pwidechar(FileName),STGM_READ or
      STGM_SHARE_DENY_WRITE,STGFMT_ANY,0,nil,nil,@IID_IPropertySetStorage,Stg);
  if Result=S_OK then begin
    PropSetStg:=Stg as IPropertySetStorage;
    Result:=PropSetStg.Enum(PropSetEnum);
    if Result=S_OK then begin
      n:=0;
      hr:=PropSetEnum.Next(1,PropSetStat,@i);
      while (hr=S_OK) and (i=1) do begin
        inc(n); SetLength(Summary,n);
        with Summary[n-1] do begin
          FormatId:=PropSetStat.fmtid;
          PropArray:=nil;
          k:=0;
          Result:=PropSetStg.Open(FormatId,STGM_READ or STGM_SHARE_EXCLUSIVE,PropStg);
          if (Result=S_OK) and (PropStg.Enum(PropEnum)=S_OK) then begin
            hr:=PropEnum.Next(1,PropStat,@i);
            while (hr=S_OK) and (i=1) do begin
              inc(k);
              SetLength(PropArray,k);
              PropSpec.ulKind:=PRSPEC_PROPID;
              PropSpec.propid:=PropStat.propid;
              with PropArray[k-1] do begin
                PropId:=PropStat.propid;
                if PropStg.ReadMultiple(1,@PropSpec,@pv)=S_OK then PropVariant:=pv
                else PropVariant.vt:=VT_Empty;
//                 and (PropVariant.vt=VT_LPSTR) and (PropVariant.pszVal <> nil) then
//                  InfoStr:=PropVariant.pszVal
//                else InfoStr:='';
                end;
              hr:=PropEnum.Next(1,PropStat,@i);
              end;
            end
          else begin
            dec(n); SetLength(Summary,n);
            end;
          end;
        hr:=PropSetEnum.Next(1,PropSetStat,@i);
        end;
      end;
    end;
  end;

function SetFileSummary(const FileName: WideString; const Summary : TFileSummary) : HResult;
var
  hr          : HRESULT;
  Stg         : IStorage;
  PropSetStg  : IPropertySetStorage;
  PropStg     : IPropertyStorage;
  PropSpec    : array of TPropSpec;
  PropVar     : array of TPropVariant;
  k,n         : integer;
begin
  if High(Summary)<0 then Result:=S_OK  // nothing to do
  else begin
    Result:=S_OK;
    // for 9x and NT4 users
    if(not Win2k) or (@StgOpenStorageEx=nil) then
      Result:=StgOpenStorage(pwidechar(FileName),nil,STGM_READWRITE or STGM_SHARE_EXCLUSIVE,nil,0,Stg)
    // for 2000, XP and higher
    else if(@StgOpenStorageEx <> nil) then
      Result:=StgOpenStorageEx(pwidechar(FileName),STGM_READWRITE or STGM_SHARE_EXCLUSIVE,
        STGFMT_FILE,0,nil,nil,@IID_IPropertySetStorage,Stg);
    if Result=S_OK then begin
      PropSetStg:=Stg as IPropertySetStorage;
      for n:=0 to High(Summary) do with Summary[n] do if Result=S_OK then begin
        // try to open an existing Property set
        hr:=PropSetStg.Open(FormatId,STGM_READWRITE or STGM_SHARE_EXCLUSIVE,PropStg);
        // create a new set
        if(hr<>S_OK) then hr:=PropSetStg.Create(FormatId,FormatId,
            PROPSETFLAG_DEFAULT,STGM_CREATE or STGM_READWRITE or STGM_SHARE_EXCLUSIVE,PropStg);
        if (hr=S_OK) then begin
          SetLength(PropSpec,length(PropArray));
          SetLength(PropVar,length(PropArray));
          for k:=0 to High(PropArray) do with PropArray[k] do begin
            PropSpec[k].ulKind:=PRSPEC_PROPID;
            PropSpec[k].propid:=PropId;
            PropVar[k]:=PropVariant;
            end;
          if (PropStg.WriteMultiple(length(PropSpec),@PropSpec[0],@PropVar[0],2)=S_OK) then
            Result:=PropStg.Commit(STGC_DEFAULT);
          end;
        end;
      end;
    end;
  end;

function GetFilePropertySet(const FileName: WideString; FormatId : TGUID) : string;
var
  hr          : HRESULT;
  Stg         : IStorage;
  PropSetStg  : IPropertySetStorage;
  PropEnum    : IEnumSTATPROPSTG;
  PropStg     : IPropertyStorage;
  PropSpec    : TPropSpec;
  PropStat    : TSTATPROPSTG;
  PropVariant : TPropVariant;
  i           : integer;
begin
  Result:='';
  hr:=S_FALSE;

  // for 9x and NT4 users
  if(not Win2k) or (@StgOpenStorageEx=nil) then
    hr:=StgOpenStorage(pwidechar(FileName),nil,STGM_READ or
      STGM_SHARE_DENY_WRITE,nil,0,Stg)
  // for 2000, XP and higher
  else if(@StgOpenStorageEx <> nil) then
    hr:=StgOpenStorageEx(pwidechar(FileName),STGM_READ or
      STGM_SHARE_DENY_WRITE,STGFMT_ANY,0,nil,nil,@IID_IPropertySetStorage,Stg);

  if hr=S_OK then begin
    PropSetStg:=Stg as IPropertySetStorage;
    if PropSetStg.Open(FormatId,STGM_READ or STGM_SHARE_EXCLUSIVE,PropStg)=S_OK then begin
      if PropStg.Enum(PropEnum)=S_OK then begin
        hr:=PropEnum.Next(1,PropStat,@i);
        while (hr=S_OK) and (i=1) do begin
          PropSpec.ulKind:=PRSPEC_PROPID;
          PropSpec.propid:=PropStat.propid;
          if (PropStg.ReadMultiple(1,@PropSpec,@PropVariant)=S_OK) and
                  (PropVariant.vt=VT_LPSTR) and (PropVariant.pszVal <> nil) then
            Result:=Result+FilePropertyToString(FormatId,PropSpec.Propid)+#9+PropVariant.pszVal+#13;
          hr:=PropEnum.Next(1,PropStat,@i);
          end;
        end;
      end;
    end;
  end;

function GetFileProperty(const FileName: WideString;
                         FormatId : TGUID; PropId : cardinal): string;
var
  hr          : HRESULT;
  Stg         : IStorage;
  PropSetStg  : IPropertySetStorage;
  PropStg     : IPropertyStorage;
  PropSpec    : TPropSpec;
  PropVariant : TPropVariant;
begin
  Result     :='';
  hr         :=S_FALSE;

  // for 9x and NT4 users
  if(not Win2k) or (@StgOpenStorageEx=nil) then
    hr:=StgOpenStorage(pwidechar(FileName),nil,STGM_READ or
      STGM_SHARE_DENY_WRITE,nil,0,Stg)
  // for 2000, XP and higher
  else if(@StgOpenStorageEx <> nil) then
    hr:=StgOpenStorageEx(pwidechar(FileName),STGM_READ or
      STGM_SHARE_DENY_WRITE,STGFMT_ANY,0,nil,nil,@IID_IPropertySetStorage,Stg);

  if(hr=S_OK) then begin
    PropSetStg:=Stg as IPropertySetStorage;
    if(PropSetStg.Open(FormatId,STGM_READ or STGM_SHARE_EXCLUSIVE,PropStg)=S_OK) then begin
      PropSpec.ulKind:=PRSPEC_PROPID;
      PropSpec.propid:=PropId;

      if(PropStg.ReadMultiple(1,@PropSpec,@PropVariant)=S_OK) and
        (PropVariant.vt=VT_LPSTR) and (PropVariant.pszVal <> nil) then
      Result:=PropVariant.pszVal;
    end;
  end;
end;

{ ------------------------------------------------------------------- }
// This routine only supports pathnames < MAX_PATH (=260)
// Reason: the called function "StgOpenStorageEx" does not support UNC paths
function CopyFileStorage (const SourceName,DestName : string) : integer;
var
  Summary : TFileSummary;
begin
  Result:=GetFileSummary(SourceName,Summary);
  if Result=S_OK then begin
    Result:=SetFileSummary(DestName,Summary);
    end;
  end;

var
  os : TOSVersionInfo;

initialization
  // get OS
  ZeroMemory(@os,sizeof(os));
  os.dwOSVersionInfoSize:=sizeof(os);
  // is it Windows 2000 or higher?
  Win2k:=(GetVersionEx(os)) and (os.dwMajorVersion >= 5) and (os.dwPlatformId=VER_PLATFORM_WIN32_NT);
  CoInitialize(nil);
  ole32dll:=FpuSafeLoadLibrary(szOleDll);
  if (ole32dll<>0) then begin
    StgOpenStorageEx:=GetProcAddress(ole32dll,'StgOpenStorageEx');
    if(@StgOpenStorageEx=nil) then begin
      FreeLibrary(ole32dll);
      ole32dll:=0;
      end;
    end;
finalization
  if(ole32dll<>0) then FreeLibrary(ole32dll);
  CoUninitialize;
end.
